__author__ = 'Lukas Leufen'
__date__ = '2021-11-18'

import pytest
import inspect
import numpy as np
import xarray as xr
import pandas as pd

from mlair.helpers.filter import ClimateFIRFilter, filter_width_kzf, firwin_kzf, omega_null_kzf, fir_filter_convolve


class TestClimateFIRFilter:

    @pytest.fixture
    def var_dim(self):
        return "variables"

    @pytest.fixture
    def time_dim(self):
        return "datetime"

    @pytest.fixture
    def data(self):
        pos = np.linspace(0, 4, num=100)
        return np.cos(pos * np.pi)

    @pytest.fixture
    def xr_array(self, data, time_dim):
        start = np.datetime64("2010-01-01 00:00")
        time_index = [start + np.timedelta64(h, "h") for h in range(len(data))]
        array = xr.DataArray(data.reshape(len(data), 1), dims=[time_dim, "station"],
                             coords={time_dim: time_index, "station": ["DE266X"]})
        return array

    @pytest.fixture
    def xr_array_long(self, data, time_dim):
        start = np.datetime64("2010-01-01 00:00")
        time_index = [start + np.timedelta64(175 * h, "h") for h in range(len(data))]
        array = xr.DataArray(data.reshape(len(data), 1), dims=[time_dim, "station"],
                             coords={time_dim: time_index, "station": ["DE266X"]})
        return array

    @pytest.fixture
    def xr_array_long_with_var(self, data, time_dim, var_dim):
        start = np.datetime64("2010-01-01 00:00")
        time_index = [start + np.timedelta64(175 * h, "h") for h in range(len(data))]
        array = xr.DataArray(data.reshape(*data.shape, 1), dims=[time_dim, "station"],
                             coords={time_dim: time_index, "station": ["DE266X"]})
        array = array.resample({time_dim: "1H"}).interpolate()
        new_data = xr.concat([array,
                              array + np.sin(np.arange(array.shape[0]) * 2 * np.pi / 24).reshape(*array.shape),
                              array + np.random.random(size=array.shape),
                              array * np.random.random(size=array.shape)],
                             dim=pd.Index(["o3", "temp", "wind", "sun"], name=var_dim))
        return new_data

    def test_combine_observation_and_apriori_no_new_dim(self, xr_array, time_dim):
        obj = object.__new__(ClimateFIRFilter)
        apriori = xr.ones_like(xr_array)
        res = obj.combine_observation_and_apriori(xr_array, apriori, time_dim, "window", 20, 10)
        assert res.coords[time_dim].values[0] == xr_array.coords[time_dim].values[20]
        first_entry = res.sel({time_dim: res.coords[time_dim].values[0]})
        assert np.testing.assert_array_equal(first_entry.sel(window=range(-20, 1)).values, xr_array.values[:21]) is None
        assert np.testing.assert_array_equal(first_entry.sel(window=range(1, 10)).values, apriori.values[21:30]) is None

    def test_combine_observation_and_apriori_with_new_dim(self, xr_array, time_dim):
        obj = object.__new__(ClimateFIRFilter)
        apriori = xr.ones_like(xr_array)
        xr_array = obj._shift_data(xr_array, range(-20, 1), time_dim, new_dim="window")
        apriori = obj._shift_data(apriori, range(1, 10), time_dim, new_dim="window")
        res = obj.combine_observation_and_apriori(xr_array, apriori, time_dim, "window", 10, 10)
        assert res.coords[time_dim].values[0] == xr_array.coords[time_dim].values[10]
        date_pos = res.coords[time_dim].values[0]
        first_entry = res.sel({time_dim: date_pos})
        assert xr.testing.assert_equal(first_entry.sel(window=range(-10, 1)),
                                       xr_array.sel({time_dim: date_pos, "window": range(-10, 1)})) is None
        assert xr.testing.assert_equal(first_entry.sel(window=range(1, 10)), apriori.sel({time_dim: date_pos})) is None

    def test_shift_data(self, xr_array, time_dim):
        remaining_dims = set(xr_array.dims).difference([time_dim])
        obj = object.__new__(ClimateFIRFilter)
        index_values = range(-15, 1)
        res = obj._shift_data(xr_array, index_values, time_dim, new_dim="window")
        assert len(res.dims) == len(remaining_dims) + 2
        assert len(set(res.dims).difference([time_dim, "window", *remaining_dims])) == 0
        assert np.testing.assert_array_equal(res.coords["window"].values, np.arange(-15, 1)) is None
        sel = res.sel({time_dim: res.coords[time_dim].values[15]})
        assert sel.sel(window=-15).values == xr_array.sel({time_dim: xr_array.coords[time_dim].values[0]}).values
        assert sel.sel(window=0).values == xr_array.sel({time_dim: xr_array.coords[time_dim].values[15]}).values

    def test_create_index_array(self):
        obj = object.__new__(ClimateFIRFilter)
        index_name = "test_index_name"
        index_values = range(-10, 1)
        res = obj.create_index_array(index_name, index_values)
        assert len(res.dims) == 1
        assert res.dims[0] == index_name
        assert res.shape == (11,)
        assert np.testing.assert_array_equal(res.values, np.arange(-10, 1)) is None

    def test_create_tmp_dimension(self, xr_array, time_dim):
        obj = object.__new__(ClimateFIRFilter)
        res = obj._create_tmp_dimension(xr_array)
        assert res == "window"
        xr_array = xr_array.rename({time_dim: "window"})
        res = obj._create_tmp_dimension(xr_array)
        assert res == "windowwindow"
        xr_array = xr_array.rename({"window": "windowwindow"})
        res = obj._create_tmp_dimension(xr_array)
        assert res == "window"

    def test_create_tmp_dimension_iter_limit(self, xr_array, time_dim):
        obj = object.__new__(ClimateFIRFilter)
        dim_name = "window"
        xr_array = xr_array.rename({time_dim: "window"})
        for i in range(11):
            dim_name += dim_name
            xr_array = xr_array.expand_dims(dim_name, -1)
        with pytest.raises(ValueError) as e:
            obj._create_tmp_dimension(xr_array)
        assert "Could not create new dimension." in e.value.args[0]

    def test_next_order(self):
        obj = object.__new__(ClimateFIRFilter)
        res = obj._next_order([43], 15, 0, "hamming")
        assert res == 15
        res = obj._next_order([43, 13], 15, 0, ("kaiser", 10))
        assert res == 28
        res = obj._next_order([43, 13], 15, 1, "hamming")
        assert res == 15
        res = obj._next_order([128, 64, 43], None, 0, "hamming")
        assert res == 64
        res = obj._next_order([43], None, 0, "hamming")
        assert res == 0

    def test_next_order_with_kzf(self):
        obj = object.__new__(ClimateFIRFilter)
        res = obj._next_order([(15, 5), (5, 3)], None, 0, "kzf")
        assert res == 13

    def test_calculate_filter_coefficients(self):
        obj = object.__new__(ClimateFIRFilter)
        res = obj._calculate_filter_coefficients("hamming", 20, 1, 24)
        assert res.shape == (20,)
        assert np.testing.assert_almost_equal(res.sum(), 1) is None
        res = obj._calculate_filter_coefficients(("kaiser", 10), 20, 1, 24)
        assert res.shape == (20,)
        assert np.testing.assert_almost_equal(res.sum(), 1) is None
        res = obj._calculate_filter_coefficients("kzf", (5, 5), 1, 24)
        assert res.shape == (21,)
        assert np.testing.assert_almost_equal(res.sum(), 1) is None

    def test_create_monthly_mean(self, xr_array_long, time_dim):
        obj = object.__new__(ClimateFIRFilter)
        res = obj.create_monthly_mean(xr_array_long, time_dim)
        assert res.shape == (1462, 1)
        assert np.datetime64("2008-12-16") == res.coords[time_dim][0].values
        assert np.datetime64("2012-12-16") == res.coords[time_dim][-1].values
        mean_jan = xr_array_long[xr_array_long[f"{time_dim}.month"] == 1].mean()
        assert res.sel({time_dim: "2009-01-16"}) == mean_jan
        mean_jul = xr_array_long[xr_array_long[f"{time_dim}.month"] == 7].mean()
        assert res.sel({time_dim: "2009-07-16"}) == mean_jul
        assert res.sel({time_dim: "2010-06-15"}) < res.sel({time_dim: "2010-06-16"})
        assert res.sel({time_dim: "2010-06-17"}) > res.sel({time_dim: "2010-06-16"})

    def test_create_monthly_mean_sampling(self, xr_array_long, time_dim):
        obj = object.__new__(ClimateFIRFilter)
        res = obj.create_monthly_mean(xr_array_long, time_dim, sampling="1m")
        assert res.shape == (49, 1)
        res = obj.create_monthly_mean(xr_array_long, time_dim, sampling="1H")
        assert res.shape == (35065, 1)
        mean_jun = xr_array_long[xr_array_long[f"{time_dim}.month"] == 6].mean()
        assert res.sel({time_dim: "2010-06-15T00:00:00"}) == mean_jun
        assert res.sel({time_dim: "2011-06-15T00:00:00"}) == mean_jun

    def test_create_monthly_mean_sel_opts(self, xr_array_long, time_dim):
        obj = object.__new__(ClimateFIRFilter)
        sel_opts = {time_dim: slice("2010-05", "2010-08")}
        res = obj.create_monthly_mean(xr_array_long, time_dim, sel_opts=sel_opts)
        assert res.dropna(time_dim)[f"{time_dim}.month"].min() == 5
        assert res.dropna(time_dim)[f"{time_dim}.month"].max() == 8
        mean_jun_2010 = xr_array_long[xr_array_long[f"{time_dim}.month"] == 6].sel({time_dim: "2010"}).mean()
        assert res.sel({time_dim: "2010-06-15T00:00:00"}) == mean_jun_2010

    def test_compute_hourly_mean_per_month(self, xr_array_long, time_dim):
        obj = object.__new__(ClimateFIRFilter)
        xr_array_long = xr_array_long.resample({time_dim: "1H"}).interpolate()
        res = obj._compute_hourly_mean_per_month(xr_array_long, time_dim, True)
        assert len(res.keys()) == 12
        assert 6 in res.keys()
        assert np.testing.assert_almost_equal(res[12].mean(), 0) is None
        assert np.testing.assert_almost_equal(res[3].mean(), 0) is None
        assert res[8].shape == (24, 1)

    def test_compute_hourly_mean_per_month_no_anomaly(self, xr_array_long, time_dim):
        obj = object.__new__(ClimateFIRFilter)
        xr_array_long = xr_array_long.resample({time_dim: "1H"}).interpolate()
        res = obj._compute_hourly_mean_per_month(xr_array_long, time_dim, False)
        assert len(res.keys()) == 12
        assert 9 in res.keys()
        assert np.testing.assert_array_less(res[2], res[1]) is None

    def test_create_seasonal_cycle_of_hourly_mean(self, xr_array_long, time_dim):
        obj = object.__new__(ClimateFIRFilter)
        xr_array_long = xr_array_long.resample({time_dim: "1H"}).interpolate()
        monthly = obj.create_monthly_unity_array(xr_array_long, time_dim) * np.nan
        seasonal_hourly_means = obj._compute_hourly_mean_per_month(xr_array_long, time_dim, True)
        res = obj._create_seasonal_cycle_of_single_hour_mean(monthly, seasonal_hourly_means, 0, time_dim, "1h")
        assert res[f"{time_dim}.hour"].sum() == 0
        assert np.testing.assert_almost_equal(res.sel({time_dim: "2010-12-01"}), res.sel({time_dim: "2011-12-01"})) is None
        res = obj._create_seasonal_cycle_of_single_hour_mean(monthly, seasonal_hourly_means, 13, time_dim, "1h")
        assert res[f"{time_dim}.hour"].mean() == 13
        assert np.testing.assert_almost_equal(res.sel({time_dim: "2010-12-01"}), res.sel({time_dim: "2011-12-01"})) is None

    def test_create_seasonal_hourly_mean(self, xr_array_long, time_dim):
        obj = object.__new__(ClimateFIRFilter)
        xr_array_long = xr_array_long.resample({time_dim: "1H"}).interpolate()
        res = obj.create_seasonal_hourly_mean(xr_array_long, time_dim)
        assert len(set(res.dims).difference(xr_array_long.dims)) == 0
        assert res.coords[time_dim][0] < xr_array_long.coords[time_dim][0]
        assert res.coords[time_dim][-1] > xr_array_long.coords[time_dim][-1]

    def test_create_seasonal_hourly_mean_sel_opts(self, xr_array_long, time_dim):
        obj = object.__new__(ClimateFIRFilter)
        xr_array_long = xr_array_long.resample({time_dim: "1H"}).interpolate()
        sel_opts = {time_dim: slice("2010-05", "2010-08")}
        res = obj.create_seasonal_hourly_mean(xr_array_long, time_dim, sel_opts=sel_opts)
        assert res.dropna(time_dim)[f"{time_dim}.month"].min() == 5
        assert res.dropna(time_dim)[f"{time_dim}.month"].max() == 8

    def test_create_unity_array(self, xr_array, time_dim):
        obj = object.__new__(ClimateFIRFilter)
        res = obj.create_monthly_unity_array(xr_array, time_dim)
        assert np.datetime64("2008-12-16") == res.coords[time_dim][0].values
        assert np.datetime64("2011-01-16") == res.coords[time_dim][-1].values
        assert res.max() == res.min()
        assert res.max() == 1
        assert res.shape == (26, 1)
        res = obj.create_monthly_unity_array(xr_array, time_dim, extend_range=0)
        assert res.shape == (1, 1)
        assert np.datetime64("2010-01-16") == res.coords[time_dim][0].values
        res = obj.create_monthly_unity_array(xr_array, time_dim, extend_range=28)
        assert res.shape == (3, 1)

    def test_extend_apriori_at_end(self, xr_array_long, time_dim):
        obj = object.__new__(ClimateFIRFilter)
        apriori = xr.ones_like(xr_array_long).sel({time_dim: "2010"})
        res = obj.extend_apriori(xr_array_long, apriori, time_dim)
        assert res.coords[time_dim][0] == apriori.coords[time_dim][0]
        assert (res.coords[time_dim][-1] - xr_array_long.coords[time_dim][-1]) / np.timedelta64(1, "D") >= 365
        apriori = xr.ones_like(xr_array_long).sel({time_dim: slice("2010", "2011-08")})
        res = obj.extend_apriori(xr_array_long, apriori, time_dim)
        assert (res.coords[time_dim][-1] - xr_array_long.coords[time_dim][-1]) / np.timedelta64(1, "D") >= (1.5 * 365)

    def test_extend_apriori_at_start(self, xr_array_long, time_dim):
        obj = object.__new__(ClimateFIRFilter)
        apriori = xr.ones_like(xr_array_long).sel({time_dim: "2011"})
        res = obj.extend_apriori(xr_array_long.sel({time_dim: slice("2010", "2010-10")}), apriori, time_dim)
        assert (res.coords[time_dim][0] - apriori.coords[time_dim][0]) / np.timedelta64(1, "D") <= -365 * 2
        assert res.coords[time_dim][-1] == apriori.coords[time_dim][-1]
        apriori = xr.ones_like(xr_array_long).sel({time_dim: slice("2010-02", "2011")})
        res = obj.extend_apriori(xr_array_long, apriori, time_dim)
        assert (res.coords[time_dim][0] - apriori.coords[time_dim][0]) / np.timedelta64(1, "D") <= -365

    def test_get_year_interval(self, xr_array, xr_array_long, time_dim):
        obj = object.__new__(ClimateFIRFilter)
        assert obj._get_year_interval(xr_array, time_dim) == (2010, 2010)
        assert obj._get_year_interval(xr_array_long, time_dim) == (2010, 2011)

    def test_create_time_range_extend(self):
        obj = object.__new__(ClimateFIRFilter)
        res = obj._create_time_range_extend(1992, "1d", 10)
        assert isinstance(res, slice)
        assert res.start == np.datetime64("1991-12-21")
        assert res.stop == np.datetime64("1993-01-11")
        assert res.step is None
        res = obj._create_time_range_extend(1992, "1H", 24)
        assert isinstance(res, slice)
        assert res.start == np.datetime64("1991-12-30T23:00:00")
        assert res.stop == np.datetime64("1993-01-01T01:00:00")
        assert res.step is None

    def test_properties(self):
        obj = object.__new__(ClimateFIRFilter)
        obj._h = [1, 2, 3]
        obj._filtered = [4, 5, 63]
        obj._apriori_list = [10, 11, 12, 13]
        assert obj.filter_coefficients == [1, 2, 3]
        assert obj.filtered_data == [4, 5, 63]
        assert obj.apriori_data == [10, 11, 12, 13]
        assert obj.initial_apriori_data == 10

    def test_trim_data_to_minimum_length(self, xr_array, time_dim):
        obj = object.__new__(ClimateFIRFilter)
        xr_array = obj._shift_data(xr_array, range(-20, 1), time_dim, new_dim="window")
        res = obj._trim_data_to_minimum_length(xr_array, 5, "window")
        assert xr_array.shape == (21, 100, 1)
        assert res.shape == (6, 100, 1)
        res = obj._trim_data_to_minimum_length(xr_array, 30, "window")
        assert res.shape == (21, 100, 1)
        xr_array = obj._shift_data(xr_array.sel(window=0), range(-20, 5), time_dim, new_dim="window")
        res = obj._trim_data_to_minimum_length(xr_array, 5, "window", extend_length_future=2)
        assert res.shape == (8, 100, 1)

    def test_create_full_filter_result_array(self, xr_array, time_dim):
        obj = object.__new__(ClimateFIRFilter)
        xr_array_window = obj._shift_data(xr_array, range(-10, 1), time_dim, new_dim="window").dropna(time_dim)
        res = obj._create_full_filter_result_array(xr_array, xr_array_window, "window")
        assert res.dims == (*xr_array.dims, "window")
        assert res.shape == (*xr_array.shape, 11)
        res2 = obj._create_full_filter_result_array(res, xr_array_window, "window")
        assert res.dims == res2.dims
        assert res.shape == res2.shape

    def test_clim_filter(self, xr_array_long_with_var, time_dim, var_dim):
        obj = object.__new__(ClimateFIRFilter)
        filter_order = 10*24+1
        res = obj.clim_filter(xr_array_long_with_var, 24, 0.05, filter_order, sampling="1H", time_dim=time_dim,
                              var_dim=var_dim, minimum_length=24)
        assert len(res) == 5

        # check filter data properties
        assert res[0].shape == (*xr_array_long_with_var.shape, int(filter_order+1)/2 + 24 + 2)
        assert res[0].dims == (*xr_array_long_with_var.dims, "window")

        # check filter properties
        assert np.testing.assert_almost_equal(
            res[2], obj._calculate_filter_coefficients("hamming",  filter_order, 0.05, 24)) is None

        # check apriori
        apriori = obj.create_monthly_mean(xr_array_long_with_var, time_dim, sampling="1H")
        apriori = apriori.astype(xr_array_long_with_var.dtype)
        apriori = obj.extend_apriori(xr_array_long_with_var, apriori, time_dim, "1H")
        assert xr.testing.assert_equal(res[3], apriori) is None

        # check plot data format
        assert isinstance(res[4], list)
        assert isinstance(res[4][0], dict)
        keys = {"t0", "var", "filter_input", "filter_input_nc", "valid_range", "time_range", "h", "new_dim"}
        assert len(keys.symmetric_difference(res[4][0].keys())) == 0

    def test_clim_filter_kwargs(self, xr_array_long_with_var, time_dim, var_dim):
        obj = object.__new__(ClimateFIRFilter)
        filter_order = 10 * 24 + 1
        apriori = obj.create_seasonal_hourly_mean(xr_array_long_with_var, time_dim, sampling="1H", as_anomaly=False)
        apriori = apriori.astype(xr_array_long_with_var.dtype)
        apriori = obj.extend_apriori(xr_array_long_with_var, apriori, time_dim, "1H")
        plot_dates = [xr_array_long_with_var.coords[time_dim][1800].values]
        res = obj.clim_filter(xr_array_long_with_var, 24, 0.05, 10 * 24 + 1, sampling="1H", time_dim=time_dim,
                              var_dim=var_dim, new_dim="total_new_dim", window=("kaiser", 5), minimum_length=1000,
                              apriori=apriori, plot_dates=plot_dates)

        assert res[0].shape == (*xr_array_long_with_var.shape, int(10 * 24 + 1 + 1) / 2 + 1000 + 2)
        assert res[0].dims == (*xr_array_long_with_var.dims, "total_new_dim")
        assert np.testing.assert_almost_equal(
            res[2], obj._calculate_filter_coefficients(("kaiser", 5),  filter_order, 0.05, 24)) is None
        assert xr.testing.assert_equal(res[3], apriori) is None
        assert len(res[4]) == len(res[0].coords[var_dim])


class TestFirFilterConvolve:

    def test_fir_filter_convolve(self):
        data = np.cos(np.linspace(0, 4, num=100) * np.pi)
        obj = object.__new__(ClimateFIRFilter)
        h = obj._calculate_filter_coefficients("hamming", 21, 0.25, 1)
        res = fir_filter_convolve(data, h)
        assert res.shape == (100,)
        assert np.testing.assert_almost_equal(np.dot(data[40:61], h) / sum(h), res[50]) is None


class TestFirwinKzf:

    def test_firwin_kzf(self):
        res = firwin_kzf(3, 3)
        assert np.testing.assert_almost_equal(res.sum(), 1) is None
        assert res.shape == (7,)
        assert np.testing.assert_array_equal(res * (3**3), np.array([1, 3, 6, 7, 6, 3, 1])) is None


class TestFilterWidthKzf:

    def test_filter_width_kzf(self):
        assert filter_width_kzf(15, 5) == 71
        assert filter_width_kzf(3, 5) == 11


class TestOmegaNullKzf:

    def test_omega_null_kzf(self):
        assert np.testing.assert_almost_equal(omega_null_kzf(13, 3), 0.01986, decimal=5) is None
        assert np.testing.assert_almost_equal(omega_null_kzf(105, 5), 0.00192, decimal=5) is None
        assert np.testing.assert_almost_equal(omega_null_kzf(3, 5), 0.07103, decimal=5) is None

    def test_omega_null_kzf_alpha(self):
        assert np.testing.assert_almost_equal(omega_null_kzf(3, 3, alpha=1), 0, decimal=1) is None
        assert np.testing.assert_almost_equal(omega_null_kzf(3, 3, alpha=0), 0.25989, decimal=5) is None
        assert np.testing.assert_almost_equal(omega_null_kzf(3, 3), omega_null_kzf(3, 3, alpha=0.5), decimal=5) is None






